@@ -12,6 +12,7 @@
 #include <cuda.h>
 #include <stdexcept>
 #include <boost/shared_ptr.hpp>
+#include <boost/foreach.hpp>
 #include <boost/weak_ptr.hpp>
 #include <utility>
 #include <stack>
@@ -22,7 +23,7 @@
 
 
 
-//#define CUDAPP_TRACE_CUDA
+// #define CUDAPP_TRACE_CUDA
 
 
 
@@ -68,10 +69,23 @@ namespace cuda
       const char *m_routine;
       CUresult m_code;
 
+    private:
+      static std::string make_message(const char *rout, CUresult c, const char *msg)
+      {
+        std::string result = rout;
+        result += " failed: ";
+        result += curesult_to_str(c);
+        if (msg)
+        {
+          result += " - ";
+          result += msg;
+        }
+        return result;
+      }
+
     public:
-      error(const char *rout, CUresult c)
-        : std::runtime_error(
-            rout + std::string(" failed: ") + curesult_to_str(c)),
+      error(const char *rout, CUresult c, const char *msg=0)
+        : std::runtime_error(make_message(rout, c, msg)),
         m_routine(rout), m_code(c)
       { }
 
@@ -219,63 +233,75 @@ namespace cuda
 
 
   // context ------------------------------------------------------------------
-  class context
+  /* A word on context management: We don't let CUDA's context stack get more
+   * than one deep. CUDA only supports pushing floating contexts. We may wish
+   * to push contexts that are already active at a deeper stack level, so we
+   * maintain all contexts floating other than the top one.
+   */
+  class context : boost::noncopyable
   {
     private:
       CUcontext m_context;
       bool m_valid;
+      unsigned m_use_count;
       typedef std::stack<boost::weak_ptr<context>,
         std::vector<boost::weak_ptr<context> > > context_stack_t;
       
       static context_stack_t m_context_stack;
 
     public:
-      context(CUcontext ctx, bool borrowed)
-        : m_context(ctx), m_valid(!borrowed)
+      context(CUcontext ctx)
+        : m_context(ctx), m_valid(true), m_use_count(1)
+      { }
+
+
+      ~context()
       { 
-        if (borrowed)
+        if (m_valid)
         {
-          CUDAPP_CALL_GUARDED(cuCtxAttach, (&m_context, 0));
-          m_valid = true;
+          if (m_use_count)
+            std::cerr 
+              << "-----------------------------------------------------------" << std::endl
+              << "PyCuda WARNING: I'm being asked to destroy a " << std::endl
+              << "context that's part of the current context stack." << std::endl
+              << "-----------------------------------------------------------" << std::endl
+              << "I will pick the next lower active context from the" << std::endl
+              << "context stack. Since this choice is happening" << std::endl
+              << "at an unspecified point in time, your code" << std::endl
+              << "may be making false assumptions about which" << std::endl
+              << "context is active at what point." << std::endl
+              << "Call Context.pop() to avoid this warning." << std::endl
+              << "-----------------------------------------------------------" << std::endl
+              << "If Python is terminating abnormally (eg. exiting upon an" << std::endl
+              << "unhandled exception), you may ignore this." << std::endl
+              << "-----------------------------------------------------------" << std::endl;
+          detach();
         }
       }
 
-      context(context const &src)
-        : m_context(src.m_context), m_valid(false)
-      { 
-        CUDAPP_CALL_GUARDED(cuCtxAttach, (&m_context, 0));
-        m_valid = true;
-      }
-
-      context &operator=(const context &src)
-      {
-        detach();
-        m_context = src.m_context;
-        CUDAPP_CALL_GUARDED(cuCtxAttach, (&m_context, 0));
-        m_valid = true;
-      }
-
-      ~context()
-      { detach(); }
+      CUcontext data() const
+      { return m_context; }
 
       void detach()
       {
         if (m_valid)
         {
-          CUDAPP_CALL_GUARDED(cuCtxDetach, (m_context));
-          m_valid = false;
-          m_context_stack.pop();
-        }
-      }
+          if (current_context().get() == this)
+          {
+            CUDAPP_CALL_GUARDED(cuCtxDetach, (m_context));
+          }
+          else
+            CUDAPP_CALL_GUARDED(cuCtxDestroy, (m_context));
 
-#if CUDA_VERSION >= 2000
-      void pop()
-      { 
-        CUcontext popped;
-        CUDAPP_CALL_GUARDED(cuCtxPopCurrent, (&popped)); 
-        if (popped != m_context)
-          throw std::runtime_error("popped the wrong context");
-        m_context_stack.pop();
+          m_valid = false;
+
+          boost::shared_ptr<context> new_active = current_context(this);
+          if (new_active.get())
+            CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (new_active->m_context)); 
+        }
+        else
+          throw error("context::detach", CUDA_ERROR_INVALID_CONTEXT,
+              "cannot detach from invalid context");
       }
 
       static device get_device()
@@ -284,14 +310,54 @@ namespace cuda
         CUDAPP_CALL_GUARDED(cuCtxGetDevice, (&dev)); 
         return device(dev);
       }
+
+#if CUDA_VERSION >= 2000
+
+      static void prepare_context_switch()
+      {
+        if (context::m_context_stack.size())
+        {
+          CUcontext popped;
+          CUDAPP_CALL_GUARDED(cuCtxPopCurrent, (&popped)); 
+        }
+      }
+
+      void pop()
+      { 
+        prepare_context_switch();
+        m_context_stack.pop();
+        --m_use_count;
+
+        boost::shared_ptr<context> current = current_context();
+        if (current)
+          CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (current_context()->m_context)); 
+      }
+#else
+      static void prepare_context_switch() { }
 #endif
 
       static void synchronize()
       { CUDAPP_CALL_GUARDED_THREADED(cuCtxSynchronize, ()); }
 
-      static boost::shared_ptr<context> current_context()
+      static boost::shared_ptr<context> current_context(context *except=0)
       {
-        return boost::shared_ptr<context>(m_context_stack.top());
+        while (true)
+        {
+          if (m_context_stack.size() == 0)
+            return boost::shared_ptr<context>();
+          boost::weak_ptr<context> result(m_context_stack.top());
+          if (!result.expired() && result.lock().get() != except)
+          {
+            // good, weak pointer didn't expire
+            // (treating except as expired weak pointer)
+            return result.lock();
+          }
+          else
+          {
+            // weak pointer invalidated, pop it and try again.
+            m_context_stack.pop();
+          }
+        }
       }
 
       friend class device;
@@ -304,36 +370,33 @@ namespace cuda
   inline
   boost::shared_ptr<context> device::make_context(unsigned int flags)
   {
+    context::prepare_context_switch();
+
     CUcontext ctx;
     CUDAPP_CALL_GUARDED(cuCtxCreate, (&ctx, flags, m_device));
-    boost::shared_ptr<context> result(new context(ctx, false));
+    boost::shared_ptr<context> result(new context(ctx));
     context::m_context_stack.push(result);
     return result;
   }
 
+
+
+
 #if CUDA_VERSION >= 2000
   inline
   void context_push(boost::shared_ptr<context> ctx)
   { 
+    context::prepare_context_switch();
+
     CUDAPP_CALL_GUARDED(cuCtxPushCurrent, (ctx->m_context)); 
     context::m_context_stack.push(ctx);
+    ++ctx->m_use_count;
   }
 #endif
 
 
 
 
-  class context_dependent
-  {
-    private:
-      boost::shared_ptr<context> m_ward_context;
-
-    public:
-      context_dependent()
-        : m_ward_context(context::current_context())
-      { }
-  };
-
   class explicit_context_dependent
   {
     private:
@@ -343,11 +406,62 @@ namespace cuda
       void acquire_context()
       {
         m_ward_context = context::current_context();
+        if (m_ward_context.get() == 0)
+          throw error("explicit_context_dependent",
+              CUDA_ERROR_INVALID_CONTEXT,
+              "no currently active context?");
       }
 
       void release_context()
       {
-        m_ward_context = boost::shared_ptr<context>();
+        m_ward_context.reset();
+      }
+
+      boost::shared_ptr<context> get_context()
+      {
+        return m_ward_context;
+      }
+  };
+
+  class context_dependent : public explicit_context_dependent
+  {
+    private:
+      boost::shared_ptr<context> m_ward_context;
+
+    public:
+      context_dependent()
+      { acquire_context(); }
+  };
+
+
+  class scoped_context_activation
+  {
+    private:
+      boost::shared_ptr<context> m_context;
+      bool m_did_switch;
+
+    public:
+      scoped_context_activation(boost::shared_ptr<context> ctx)
+        : m_context(ctx)
+      { 
+        m_did_switch = context::current_context() != m_context;
+        if (m_did_switch)
+        {
+#if CUDA_VERSION >= 2000
+          context_push(m_context);
+#else
+          throw cuda::error("scoped_context_activation", CUDA_ERROR_INVALID_CONTEXT,
+              "not available in CUDA < 2.0");
+#endif
+        }
+      }
+
+      ~scoped_context_activation()
+      {
+#if CUDA_VERSION >= 2000
+        if (m_did_switch)
+          m_context->pop();
+#endif
       }
 
   };
@@ -363,7 +477,10 @@ namespace cuda
       { CUDAPP_CALL_GUARDED(cuStreamCreate, (&m_stream, flags)); }
 
       ~stream()
-      { CUDAPP_CALL_GUARDED(cuStreamDestroy, (m_stream)); }
+      { 
+        scoped_context_activation ca(get_context());
+        CUDAPP_CALL_GUARDED(cuStreamDestroy, (m_stream)); 
+      }
 
       void synchronize()
       { CUDAPP_CALL_GUARDED_THREADED(cuStreamSynchronize, (m_stream)); }
@@ -415,10 +532,19 @@ namespace cuda
       { }
 
       ~array()
-      { 
+      { free(); }
+
+      void free()
+      {
         if (m_managed)
         {
-          CUDAPP_CALL_GUARDED(cuArrayDestroy, (m_array)); 
+          {
+            scoped_context_activation ca(get_context());
+            CUDAPP_CALL_GUARDED(cuArrayDestroy, (m_array)); 
+          }
+
+          m_managed = false;
+          release_context();
         }
       }
 
@@ -569,6 +695,7 @@ namespace cuda
 
       ~module()
       {
+        scoped_context_activation ca(get_context());
         CUDAPP_CALL_GUARDED(cuModuleUnload, (m_module));
       }
 
@@ -698,7 +825,9 @@ namespace cuda
       {
         if (m_valid)
         {
+          scoped_context_activation ca(get_context());
           mem_free(m_devptr);
+          release_context();
           m_valid = false;
         }
         else
@@ -708,19 +837,13 @@ namespace cuda
       ~device_allocation()
       {
         if (m_valid)
-          mem_free(m_devptr);
+          free();
       }
       
       operator CUdeviceptr() const
       { return m_devptr; }
   };
 
-  inline 
-  device_allocation *make_device_allocation(unsigned long bytes)
-  {
-    return new device_allocation(mem_alloc(bytes));
-  }
-
   inline unsigned int mem_alloc_pitch(
       std::auto_ptr<device_allocation> &da,
         unsigned int width, unsigned int height, unsigned int access_size)
@@ -762,11 +885,17 @@ namespace cuda
 
 
   // structured memcpy --------------------------------------------------------
+#if PY_VERSION_HEX >= 0x02050000
+  typedef Py_ssize_t PYCUDA_BUFFER_SIZE_T;
+#else
+  typedef int PYCUDA_BUFFER_SIZE_T;
+#endif
+
 #define MEMCPY_SETTERS \
     void set_src_host(py::object buf_py) \
     { \
       srcMemoryType = CU_MEMORYTYPE_HOST; \
-      Py_ssize_t len; \
+      PYCUDA_BUFFER_SIZE_T len; \
       if (PyObject_AsReadBuffer(buf_py.ptr(), &srcHost, &len)) \
         throw py::error_already_set(); \
     } \
@@ -786,7 +915,7 @@ namespace cuda
     void set_dst_host(py::object buf_py) \
     { \
       dstMemoryType = CU_MEMORYTYPE_HOST; \
-      Py_ssize_t len; \
+      PYCUDA_BUFFER_SIZE_T len; \
       if (PyObject_AsWriteBuffer(buf_py.ptr(), &dstHost, &len)) \
         throw py::error_already_set(); \
     } \
@@ -843,10 +972,12 @@ namespace cuda
       srcXInBytes = 0;
       srcY = 0;
       srcZ = 0;
+      srcLOD = 0;
 
       dstXInBytes = 0;
       dstY = 0;
       dstZ = 0;
+      dstLOD = 0;
     }
 
     MEMCPY_SETTERS;
@@ -863,6 +994,21 @@ namespace cuda
 
 
   // host memory --------------------------------------------------------------
+  inline void *mem_alloc_host(unsigned int size)
+  {
+    void *m_data;
+    CUDAPP_CALL_GUARDED(cuMemAllocHost, (&m_data, size));
+    return m_data;
+  }
+
+  inline void mem_free_host(void *ptr)
+  {
+    CUDAPP_CALL_GUARDED(cuMemFreeHost, (ptr));
+  }
+
+
+
+
   struct host_allocation : public boost::noncopyable
   {
     private:
@@ -870,11 +1016,20 @@ namespace cuda
 
     public:
       host_allocation(unsigned bytesize)
-        : m_data(0)
-      { CUDAPP_CALL_GUARDED(cuMemAllocHost, (&m_data, bytesize)); }
+        : m_data(mem_alloc_host(bytesize))
+      { }
 
       ~host_allocation()
-      { CUDAPP_CALL_GUARDED(cuMemFreeHost, (m_data)); }
+      { free(); }
+
+      void free()
+      {
+        if (m_data)
+        {
+          mem_free_host(m_data); 
+          m_data = 0;
+        }
+      }
       
       void *data()
       { return m_data; }
@@ -894,7 +1049,10 @@ namespace cuda
       { CUDAPP_CALL_GUARDED(cuEventCreate, (&m_event, flags)); }
 
       ~event()
-      { CUDAPP_CALL_GUARDED(cuEventDestroy, (m_event)); }
+      { 
+        scoped_context_activation ca(get_context());
+        CUDAPP_CALL_GUARDED(cuEventDestroy, (m_event)); 
+      }
 
       void record()
       { CUDAPP_CALL_GUARDED(cuEventRecord, (m_event, 0)); }
