@@ -8,3 +8,5 @@ assert cuda.Device.count() >= 1
 device = pycuda.tools.get_default_device()
 context = device.make_context()
 
+import atexit
+atexit.register(context.pop)
