@@ -409,7 +409,11 @@ def from_device_like(devptr, other_ary):
 @memoize
 def _get_nvcc_version(nvcc):
     from subprocess import Popen, PIPE
-    return Popen([nvcc, "--version"], stdout=PIPE).communicate()[0]
+    try:
+        return Popen([nvcc, "--version"], stdout=PIPE).communicate()[0]
+    except OSError, e:
+        raise OSError, "%s was not found (is it on the PATH?) [%s]" % (
+                nvcc, str(e))
 
 
 
@@ -438,7 +442,10 @@ def _do_compile(source, options, keep, nvcc, cache_dir):
     file_dir = mkdtemp()
     file_root = "kernel"
 
-    outf = open(join(file_dir, file_root + ".cu"), "w")
+    cu_file_name = file_root + ".cu"
+    cu_file_path = join(file_dir, cu_file_name)
+
+    outf = open(cu_file_path, "w")
     outf.write(str(source))
     outf.close()
 
@@ -452,14 +459,14 @@ def _do_compile(source, options, keep, nvcc, cache_dir):
     try:
         result = call([nvcc, "--cubin"]
                 + options
-                + [file_root + ".cu"],
+                + [cu_file_name],
             cwd=file_dir)
     except OSError, e:
         raise OSError, "%s was not found (is it on the PATH?) [%s]" % (
                 nvcc, str(e))
 
     if result != 0:
-        raise CompileError, "module compilation failed"
+        raise CompileError, "nvcc compilation of %s failed" % cu_file_path
 
     cubin = open(join(file_dir, file_root + ".cubin"), "r").read()
 
@@ -498,8 +505,10 @@ class SourceModule(object):
 
         if cache_dir is None:
             from os.path import expanduser, join, exists
-            cache_dir = join(expanduser("~"), 
-                    ".pycuda-compiler-cache")
+            import os
+            from tempfile import gettempdir
+            cache_dir = join(gettempdir(), 
+                    "pycuda-compiler-cache-v1-uid%s" % os.getuid())
 
             if not exists(cache_dir):
                 from os import mkdir
