@@ -1,6 +1,6 @@
 from __future__ import division
 import numpy
-import pycuda._kernel as _kernel
+import pycuda.elementwise as elementwise
 import random as random
 from pytools import memoize
 from pytools.diskdict import DiskDict
@@ -61,7 +61,8 @@ class GPUArray(object):
         s = 1
         for dim in shape:
             s *= dim
-        self.size = s
+        self.mem_size = self.size = s
+        self.nbytes = self.dtype.itemsize * self.size
 
         self.allocator = allocator
         if self.size:
@@ -70,12 +71,12 @@ class GPUArray(object):
             self.gpudata = None
         self.stream = stream
 
-        self._grid, self._block = splay(self.size)
+        self._grid, self._block = splay(self.mem_size)
 
     @classmethod
     def compile_kernels(cls):
         # useful for benchmarking
-        _kernel._compile_kernels(cls)
+        elementwise._compile_kernels(cls)
 
     def set(self, ary, stream=None):
         assert ary.size == self.size
@@ -110,17 +111,17 @@ class GPUArray(object):
         assert self.shape == other.shape
         assert self.dtype == other.dtype
 
-        func = _kernel.get_axpbyz_kernel()
+        func = elementwise.get_axpbyz_kernel()
         func.set_block_shape(*self._block)
 
         if add_timer is not None:
             add_timer(3*self.size, func.prepared_timed_call(self._grid, 
                 selffac, self.gpudata, otherfac, other.gpudata, 
-                out.gpudata, self.size))
+                out.gpudata, self.mem_size))
         else:
             func.prepared_async_call(self._grid, self.stream,
                     selffac, self.gpudata, otherfac, other.gpudata, 
-                    out.gpudata, self.size)
+                    out.gpudata, self.mem_size)
 
         return out
 
@@ -128,11 +129,11 @@ class GPUArray(object):
         """Compute ``out = selffac * self + other``, where `other` is a scalar."""
         assert self.dtype == numpy.float32
 
-        func = _kernel.get_axpbz_kernel()
+        func = elementwise.get_axpbz_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
                 selffac, self.gpudata,
-                other, out.gpudata, self.size)
+                other, out.gpudata, self.mem_size)
 
         return out
 
@@ -140,11 +141,11 @@ class GPUArray(object):
         assert self.dtype == numpy.float32
         assert self.dtype == numpy.float32
 
-        func = _kernel.get_multiply_kernel()
+        func = elementwise.get_multiply_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
                 self.gpudata, other.gpudata,
-                out.gpudata, self.size)
+                out.gpudata, self.mem_size)
 
         return out
 
@@ -156,11 +157,11 @@ class GPUArray(object):
 
         assert self.dtype == numpy.float32
 
-        func = _kernel.get_rdivide_scalar_kernel()
+        func = elementwise.get_rdivide_elwise_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
                 self.gpudata, other,
-                out.gpudata, self.size)
+                out.gpudata, self.mem_size)
 
         return out
 
@@ -171,11 +172,11 @@ class GPUArray(object):
         assert self.shape == other.shape
         assert self.dtype == other.dtype
 
-        func = _kernel.get_divide_kernel()
+        func = elementwise.get_divide_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
                 self.gpudata, other.gpudata,
-                out.gpudata, self.size)
+                out.gpudata, self.mem_size)
 
         return out
 
@@ -192,50 +193,42 @@ class GPUArray(object):
     def __add__(self, other):
         """Add an array with an array or an array with a scalar."""
 
-        if isinstance(other, (int, float, complex)):
+        if isinstance(other, GPUArray):
+            # add another vector
+            result = self._new_like_me()
+            return self._axpbyz(1, other, 1, result)
+        else:
             # add a scalar
             if other == 0:
                 return self
             else:
                 result = self._new_like_me()
                 return self._axpbz(1, other, result)
-        else:
-            # add another vector
-            result = self._new_like_me()
-            return self._axpbyz(1, other, 1, result)
 
     __radd__ = __add__
 
     def __sub__(self, other):
         """Substract an array from an array or a scalar from an array."""
 
-        if isinstance(other, (int, float, complex)):
-            # if array - 0 than just return the array since its the same anyway
-
+        if isinstance(other, GPUArray):
+            result = self._new_like_me()
+            return self._axpbyz(1, other, -1, result)
+        else:
             if other == 0:
                 return self
             else:
                 # create a new array for the result
                 result = self._new_like_me()
                 return self._axpbz(1, -other, result)
-        else:
-            result = self._new_like_me()
-            return self._axpbyz(1, other, -1, result)
 
     def __rsub__(self,other):
         """Substracts an array by a scalar or an array:: 
 
            x = n - self
         """
-        assert isinstance(other, (int, float, complex))
-
-        # if array - 0 than just return the array since its the same anyway
-        if other == 0:
-            return self
-        else:
-            # create a new array for the result
-            result = self._new_like_me()
-            return self._axpbz(-1, other, result)
+        # other must be a scalar
+        result = self._new_like_me()
+        return self._axpbz(-1, other, result)
 
     def __iadd__(self, other):
         return self._axpbyz(1, other, 1, self)
@@ -249,10 +242,10 @@ class GPUArray(object):
 
     def __mul__(self, other):
         result = self._new_like_me()
-        if isinstance(other, (int, float, complex)):
-            return self._axpbz(other, 0, result)
-        else:
+        if isinstance(other, GPUArray):
             return self._elwise_multiply(other, result)
+        else:
+            return self._axpbz(other, 0, result)
 
     def __rmul__(self, scalar):
         result = self._new_like_me()
@@ -266,17 +259,18 @@ class GPUArray(object):
 
            x = self / n
         """
-        if isinstance(other, (int, float, complex)):
-            # if array - 0 than just return the array since its the same anyway
-            if other == 0:
+        if isinstance(other, GPUArray):
+            result = self._new_like_me()
+            return self._div(other, result)
+        else:
+            if other == 1:
                 return self
             else:
                 # create a new array for the result
                 result = self._new_like_me()
                 return self._axpbz(1/other, 0, result)
-        else:
-            result = self._new_like_me()
-            return self._div(other, result)
+
+    __truediv__ = __div__
 
     def __rdiv__(self,other):
         """Divides an array by a scalar or an array::
@@ -284,34 +278,34 @@ class GPUArray(object):
            x = n / self
         """
 
-        if isinstance(other, (int, float, complex)):
-            # if array - 0 than just return the array since its the same anyway
-            if other == 0:
+        if isinstance(other, GPUArray):
+            result = self._new_like_me()
+
+            func = elementwise.get_divide_kernel()
+            func.set_block_shape(*self._block)
+            func.prepared_async_call(self._grid, self.stream,
+                    other.gpudata, self.gpudata, out.gpudata, 
+                    self.mem_size)
+
+            return result
+        else:
+            if other == 1:
                 return self
             else:
                 # create a new array for the result
                 result = self._new_like_me()
                 return self._rdiv_scalar(other, result)
-        else:
-            result = self._new_like_me()
-
-            func = _kernel.get_divide_kernel()
-            func.set_block_shape(*self._block)
-            func.prepared_async_call(self._grid, self.stream,
-                    other.gpudata, self.gpudata, out.gpudata, 
-                    self.size)
-
-            return result
 
+    __rtruediv__ = __div__
 
     def fill(self, value):
         """fills the array with the specified value"""
         assert self.dtype == numpy.float32
 
-        func = _kernel.get_fill_kernel()
+        func = elementwise.get_fill_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
-                value, self.gpudata, self.size)
+                value, self.gpudata, self.mem_size)
 
         return self
 
@@ -331,10 +325,10 @@ class GPUArray(object):
 
         result = GPUArray(self.shape, self.dtype)
 
-        func = _kernel.get_unary_func_kernel("fabs")
+        func = elementwise.get_unary_func_kernel("fabs")
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
-                self.gpudata,result.gpudata, self.size)
+                self.gpudata,result.gpudata, self.mem_size)
 
         return result
 
@@ -349,25 +343,25 @@ class GPUArray(object):
         """
         result = GPUArray(self.shape, self.dtype)
 
-        if isinstance(other, (int, float, complex)):
-            func = _kernel.get_pow_kernel()
-            func.set_block_shape(*self._block)
-            func.prepared_async_call(self._grid, self.stream,
-                    other, self.gpudata, result.gpudata,
-                    self.size)
-
-            return result
-        else:
+        if isinstance(other, GPUArray):
             assert self.shape == other.shape
             assert self.dtype == other.dtype
 
-            func = _kernel.get_pow_array_kernel()
+            func = elementwise.get_pow_array_kernel()
             func.set_block_shape(*self._block)
             func.prepared_async_call(self._grid, self.stream,
                     self.gpudata, other.gpudata, result.gpudata,
-                    self.size)
+                    self.mem_size)
             
             return result
+        else:
+            func = elementwise.get_pow_kernel()
+            func.set_block_shape(*self._block)
+            func.prepared_async_call(self._grid, self.stream,
+                    other, self.gpudata, result.gpudata,
+                    self.mem_size)
+
+            return result
 
     def reverse(self):
         """Return this array in reversed order. The array is treated
@@ -378,11 +372,11 @@ class GPUArray(object):
         
         result = GPUArray(self.shape, self.dtype)
 
-        func = _kernel.get_reverse_kernel()
+        func = elementwise.get_reverse_kernel()
         func.set_block_shape(*self._block)
         func.prepared_async_call(self._grid, self.stream,
                 self.gpudata, result.gpudata,
-                self.size)
+                self.mem_size)
 
         return result
 
@@ -404,6 +398,20 @@ def zeros(shape, dtype, stream=None, allocator=drv.mem_alloc):
     result.fill(0)
     return result
 
+def empty_like(other_ary):
+    result = GPUArray(
+            other_ary.shape, other_ary.dtype, 
+            other_ary.stream, other_ary.allocator)
+    return result
+
+def zeros_like(other_ary):
+    result = GPUArray(
+            other_ary.shape, other_ary.dtype, 
+            other_ary.stream, other_ary.allocator)
+    result.fill(0)
+    return result
+
+
 def arange(*args, **kwargs):
     """Create an array filled with numbers spaced `step` apart,
     starting from `start` and ending at `stop`.
@@ -471,7 +479,7 @@ def arange(*args, **kwargs):
   
     result = GPUArray((size,), dtype)
 
-    func = _kernel.get_arange_kernel()
+    func = elementwise.get_arange_kernel()
     func.set_block_shape(*result._block)
     func.prepared_async_call(result._grid, result.stream,
             result.gpudata, start, step, size)
