@@ -4,9 +4,9 @@
 #include <numeric>
 #include <algorithm>
 
+#include "tools.hpp"
 #include "wrap_helpers.hpp"
 #include <boost/python/stl_iterator.hpp>
-#include "numpy_init.hpp"
 
 
 
@@ -72,6 +72,11 @@ namespace
 
 
 
+  device_allocation *mem_alloc_wrap(unsigned long bytes)
+  {
+    return new device_allocation(pycuda::mem_alloc_gc(bytes));
+  }
+
   py::tuple mem_alloc_pitch_wrap(
       unsigned int width, unsigned int height, unsigned int access_size)
   {
@@ -110,7 +115,7 @@ namespace
   void function_param_setv(function &f, int offset, py::object buffer)
   { 
     const void *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsReadBuffer(buffer.ptr(), &buf, &len))
       throw py::error_already_set();
     f.param_setv(offset, const_cast<void *>(buf), len);
@@ -122,7 +127,7 @@ namespace
   module *module_from_buffer(py::object buffer)
   {
     const char *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsCharBuffer(buffer.ptr(), &buf, &len))
       throw py::error_already_set();
     CUmodule mod;
@@ -141,18 +146,6 @@ namespace
 
 
 
-  inline
-  npy_intp size_from_dims(int ndim, const npy_intp *dims)
-  {
-    if (ndim != 0)
-      return std::accumulate(dims, dims+ndim, 1, std::multiplies<npy_intp>());
-    else
-      return 1;
-  }
-
-
-
-
   py::handle<> pagelocked_empty(py::object shape, py::object dtype, py::object order_py)
   {
     PyArray_Descr *tp_descr;
@@ -167,7 +160,7 @@ namespace
 
     std::auto_ptr<host_allocation> alloc(
         new host_allocation(
-          tp_descr->elsize*size_from_dims(dims.size(), &dims.front())));
+          tp_descr->elsize*pycuda::size_from_dims(dims.size(), &dims.front())));
 
     NPY_ORDER order = PyArray_CORDER;
     PyArray_OrderConverter(order_py.ptr(), &order);
@@ -198,7 +191,7 @@ namespace
   void py_memcpy_htod(CUdeviceptr dst, py::object src, py::object stream_py)
   {
     const void *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsReadBuffer(src.ptr(), &buf, &len))
       throw py::error_already_set();
 
@@ -219,7 +212,7 @@ namespace
   void py_memcpy_dtoh(py::object dest, CUdeviceptr src, py::object stream_py)
   {
     void *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsWriteBuffer(dest.ptr(), &buf, &len))
       throw py::error_already_set();
 
@@ -240,7 +233,7 @@ namespace
   void py_memcpy_htoa(array const &ary, unsigned int index, py::object src)
   {
     const void *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsReadBuffer(src.ptr(), &buf, &len))
       throw py::error_already_set();
 
@@ -253,7 +246,7 @@ namespace
   void py_memcpy_atoh(py::object dst, array const &ary, unsigned int index)
   {
     void *buf;
-    Py_ssize_t len;
+    PYCUDA_BUFFER_SIZE_T len;
     if (PyObject_AsWriteBuffer(dst.ptr(), &buf, &len))
       throw py::error_already_set();
 
@@ -378,7 +371,7 @@ BOOST_PYTHON_MODULE(_driver)
 
   {
     typedef context cl;
-    py::class_<cl, shared_ptr<cl> >("Context", py::no_init)
+    py::class_<cl, shared_ptr<cl>, boost::noncopyable >("Context", py::no_init)
       .DEF_SIMPLE_METHOD(detach)
 
 #if CUDA_VERSION >= 2000
@@ -450,11 +443,12 @@ BOOST_PYTHON_MODULE(_driver)
   {
     typedef host_allocation cl;
     py::class_<cl, boost::noncopyable>("HostAllocation", py::no_init)
+      .DEF_SIMPLE_METHOD(free)
       ;
   }
 
   DEF_SIMPLE_FUNCTION(mem_get_info);
-  py::def("mem_alloc", make_device_allocation, 
+  py::def("mem_alloc", mem_alloc_wrap, 
       py::return_value_policy<py::manage_new_object>());
   py::def("mem_alloc_pitch", mem_alloc_pitch_wrap,
       py::args("width", "height", "access_size"));
@@ -515,7 +509,7 @@ BOOST_PYTHON_MODULE(_driver)
       .def_readwrite("width_in_bytes", &cl::WidthInBytes)
       .def_readwrite("height", &cl::Height)
 
-      .def("__call__", &cl::execute, py::arg("aligned"))
+      .def("__call__", &cl::execute, py::args("self", "aligned"))
       .def("__call__", &cl::execute_async)
       ;
   }
@@ -552,9 +546,9 @@ BOOST_PYTHON_MODULE(_driver)
 
       .def_readwrite("width_in_bytes", &cl::WidthInBytes)
       .def_readwrite("height", &cl::Height)
-      .def_readwrite("height", &cl::Depth)
+      .def_readwrite("depth", &cl::Depth)
 
-      .def("__call__", &cl::execute, py::arg("aligned"))
+      .def("__call__", &cl::execute)
       .def("__call__", &cl::execute_async)
       ;
   }
@@ -603,6 +597,7 @@ BOOST_PYTHON_MODULE(_driver)
     typedef array cl;
     py::class_<cl, shared_ptr<cl>, boost::noncopyable>
       ("Array", py::init<const CUDA_ARRAY_DESCRIPTOR &>())
+      .DEF_SIMPLE_METHOD(free)
       .DEF_SIMPLE_METHOD(get_descriptor)
 #if CUDA_VERSION >= 2000
       .def(py::init<const CUDA_ARRAY3D_DESCRIPTOR &>())
