@@ -1,3 +1,4 @@
+from __future__ import division
 import unittest
 import pycuda.autoinit
 import pycuda.driver as drv
@@ -42,6 +43,37 @@ class TestCuda(unittest.TestCase):
                 block=(400,1,1))
         self.assert_(la.norm(dest-a*b) == 0)
 
+    def test_simple_kernel_2(self):
+        mod = drv.SourceModule("""
+        __global__ void multiply_them(float *dest, float *a, float *b)
+        {
+          const int i = threadIdx.x;
+          dest[i] = a[i] * b[i];
+        }
+        """)
+
+        multiply_them = mod.get_function("multiply_them")
+
+        import numpy
+        a = numpy.random.randn(400).astype(numpy.float32)
+        b = numpy.random.randn(400).astype(numpy.float32)
+        a_gpu = drv.to_device(a)
+        b_gpu = drv.to_device(b)
+
+        dest = numpy.zeros_like(a)
+        multiply_them(
+                drv.Out(dest), a_gpu, b_gpu,
+                block=(400,1,1))
+        self.assert_(la.norm(dest-a*b) == 0)
+
+        # now try with offsets
+        dest = numpy.zeros_like(a)
+        multiply_them(
+                drv.Out(dest), numpy.intp(a_gpu)+1, b_gpu,
+                block=(399,1,1))
+
+        self.assert_(la.norm(dest-a*b) == 0)
+
     def test_streamed_kernel(self):
         # this differs from the "simple_kernel" case in that *all* computation
         # and data copying is asynchronous. Observe how this necessitates the
@@ -87,7 +119,7 @@ class TestCuda(unittest.TestCase):
         diff = ((a_g*b_g).get()-a*b)
         assert la.norm(diff) == 0
 
-    def do_not_test_cublas_mixing(self):
+    def donottest_cublas_mixing(self):
         self.test_streamed_kernel()
 
         import pycuda.blas as blas
@@ -223,6 +255,19 @@ class TestCuda(unittest.TestCase):
         assert bitlog2(0x3affe) == 17
         assert bitlog2(0xcc3affe) == 27
 
+    def test_mempool_2(self):
+        from pycuda.tools import DeviceMemoryPool as DMP
+        from random import randrange
+
+        for i in range(2000):
+            s = randrange(1<<31) >> randrange(32)
+            bin_nr = DMP.bin_number(s)
+            asize = DMP.alloc_size(bin_nr)
+
+            assert asize >= s, s
+            assert DMP.bin_number(asize) == bin_nr, s
+            assert asize < asize*(1+1/8)
+            
     def test_mempool(self):
         from pycuda.tools import bitlog2
         from pycuda.tools import DeviceMemoryPool
@@ -240,6 +285,100 @@ class TestCuda(unittest.TestCase):
                 if len(queue) > 10:
                     queue.pop(0)
 
+    def test_multi_context(self):
+        if drv.get_version() < (2,0,0):
+            return
+
+        mem_a = drv.mem_alloc(50)
+        ctx2 = pycuda.autoinit.device.make_context()
+        mem_b = drv.mem_alloc(60)
+
+        del mem_a
+        del mem_b
+        ctx2.detach()
+
+    def test_3d_texture(self):
+        # adapted from code by Nicolas Pinto
+        w = 2
+        h = 4
+        d = 8
+        shape = (w, h, d)
+
+        a = numpy.asarray(
+                numpy.random.randn(*shape),
+                dtype=numpy.float32, order="F")
+
+        descr = drv.ArrayDescriptor3D()
+        descr.width = w
+        descr.height = h
+        descr.depth = d
+        descr.format = drv.dtype_to_array_format(a.dtype)
+        descr.num_channels = 1
+        descr.flags = 0
+
+        ary = drv.Array(descr)
+
+        copy = drv.Memcpy3D()
+        copy.set_src_host(a)
+        copy.set_dst_array(ary)
+        copy.width_in_bytes = copy.src_pitch = a.strides[1]
+        copy.src_height = copy.height = h
+        copy.depth = d
+
+        copy()
+
+        mod = drv.SourceModule("""
+        texture<float, 3, cudaReadModeElementType> mtx_tex;
+
+        __global__ void copy_texture(float *dest)
+        {
+          int x = threadIdx.x;
+          int y = threadIdx.y;
+          int z = threadIdx.z;
+          int dx = blockDim.x;
+          int dy = blockDim.y;
+          int i = (z*dy + y)*dx + x;
+          dest[i] = tex3D(mtx_tex, x, y, z);
+          //dest[i] = x;
+        }
+        """)
+
+        copy_texture = mod.get_function("copy_texture")
+        mtx_tex = mod.get_texref("mtx_tex")
+
+        mtx_tex.set_array(ary)
+
+        dest = numpy.zeros(shape, dtype=numpy.float32, order="F")
+        copy_texture(drv.Out(dest), block=shape, texrefs=[mtx_tex])
+        assert la.norm(dest-a) == 0
+
+    def test_prepared_invocation(self):
+        a = numpy.random.randn(4,4).astype(numpy.float32)
+        a_gpu = drv.mem_alloc(a.size * a.dtype.itemsize)
+
+        drv.memcpy_htod(a_gpu, a)
+
+        mod = drv.SourceModule("""
+            __global__ void doublify(float *a)
+            {
+              int idx = threadIdx.x + threadIdx.y*blockDim.x;
+              a[idx] *= 2;
+            }
+            """)
+
+        func = mod.get_function("doublify")
+        func.prepare("P", (4,4,1))
+        func.prepared_call((1, 1), a_gpu)
+        a_doubled = numpy.empty_like(a)
+        drv.memcpy_dtoh(a_doubled, a_gpu)
+        assert la.norm(a_doubled-2*a) == 0
+
+        # now with offsets
+        func.prepare("P", (15,1,1))
+        a_quadrupled = numpy.empty_like(a)
+        func.prepared_call((1, 1), int(a_gpu)+a.dtype.itemsize)
+        drv.memcpy_dtoh(a_quadrupled, a_gpu)
+        assert la.norm(a_quadrupled[1:]-4*a[1:]) == 0
 
 
 
